
local VERSION = string.match("1.1.0", "^v?[%d+.]+....$") or DEV_BUILD or "Build: 2013-10-13T20:53:26Z"
local BUILD = tonumber("20131013205326") or 99999999999999
local ROOT = "Interface/Addons/AA_awsmDropdown"

-- namespaces
local Library = {
	System = {
		Name = "awsmDropdown",
		Path = ROOT,
		Version = VERSION,
		Build = BUILD,
		Author = "Noguai",
	},
	Button = {},
	DropDown = {},
}
_G.ADD = Library

local History = {}
local SupressValueUpdates = false
local SupressAutoClose = false
local SupressChange = false
local MouseoverButton = true

local Button = Library.Button
local DropDown = Library.DropDown
_G.awsmDropdown = DropDown --create another global namespace for easier usage of main functionality

-- performance locals
local zzLibrary = ZZLibrary
local zzTable, zzMath, zzString, zzClasses, zzHash, zzColors, zzStory, zzEvent, zzTimer, zzWidget, zzConfig = zzLibrary.GetLibraries()
local string = string
local math = math
local table = table
local pairs = pairs
local ipairs = ipairs
local type = type
local tostring = tostring
local tonumber = tonumber
local error = error
local assert = assert
local loadstring = loadstring
local pcall = pcall

-- internal tools
local function throw(msg, level, ...)
	for i, v in ipairs(arg) do
		msg = string.format("%s\n%d: %s", msg, i, tostring(v))
	end
	error(msg, level + 1)
end
local function catch(test, msg, ...)
	if not test then throw(msg, 2, ...) end
end
local function argn(...)
	return arg.n
end
local function println(msg, sender)
	if sender and sender.Name then
		msg = msg or "" -- for string.format
		if sender.Version then
			DEFAULT_CHAT_FRAME:AddMessage(string.format("|cFFFF0000%s (v. %s):|r %s", sender.Name, sender.Version, msg))
		else
			DEFAULT_CHAT_FRAME:AddMessage(string.format("|cFFFF0000%s:|r %s", sender.Name, msg))
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage(msg)
	end
end
local function print(...)
	for _, msg in ipairs(arg) do
		println(msg)
	end
end
local function printf(format, ...)
	local needed = 0
	for _ in format:gmatch("%%.") do
		needed = needed + 1;
	end

	if needed == argn(...) then
		DEFAULT_CHAT_FRAME:AddMessage(format:format(...), 1, 1, 1);
	else
		error("Argument number mismatch!", 2)
	end
end
local function arglist(...)
	local argNum = select('#', ...)
	local argStr = ""
	if argNum > 0 then
		argStr = "Argument list:"
		for i = 1, argNum do
			local arg = select(i, ...)
			if arg ~= nil then
				local argType = type(arg)
				local str = argType == "string" and string.format("'%s'", arg) or tostring(arg)
				argStr = string.format("%s\n%d: %s", argStr, i, str)
			end
		end
	end

	return argStr
end

local function checktype(var, types, argn, func, ...)
	local varType = type(var)
	local typesType = type(types)

	if typesType == "string" then
		if types == varType then
			return
		end
	else
		for i, type in pairs(types) do
			if type == varType then
				return
			end
		end
	end

	local argStr = arglist(...)
	local validTypeStr = typesType == "string" and types or table.concat(types, ", or ")
	throw(string.format("Invalid argument #%d for %s. (got %s; expected %s)\n%s", tonumber(argn) or 0, tostring(func), varType, validTypeStr, argStr), 3)
end

local function bug(errMsg, forceFrame)
	printc(0.6, 0.6, 0.6, "|cffff0000== SCRIPT_RUNTIIME_ERROR ==|r", errMsg)

	BugMessageFrame_AddText(errMsg)
	BUG_MESSAGE_NEW_ITEM = true
	MinimapFrameBugGartherButton:Show()

	if forceFrame then
		BugMessageFrame:Show()
	end
end

local function bugf(errMsg, ...)
	local needVarArg = 0
	local msg

	for _ in errMsg:gmatch("%%.") do
		needVarArg = needVarArg + 1;
	end

	if needVarArg == argn(...) then
		msg = errMsg:format(...)
	else
		msg = errMsg
	end

	bug(msg)
end

local CompiledScripts = {}
local function runScript(script, ...)
	local Type = type(script)
	if Type == "string" then
		if CompiledScripts[script] then
			CompiledScripts[script]()
		else
			local func, err = loadstring(script)
			assert(func, err)
			CompiledScripts[script] = func
			func()
		end
	elseif Type == "function" then
		script(...)
	end
end

---
-- @class table
-- @name DropDownButton
-- @description The information about a dropdown button (called "the buttons data"; you can also add custom fields to use them in callback functions)
-- @field Type (string) The buttons control type (Button, CheckButton, Slider, EditBox, DropDown, Label; default: Button)
-- @field Title (string) The displayed title of the button
-- @field Icon (string) The path to the icon on the button
-- @field Enabled (bool) Whether to enable or disable the button
-- @field Click (function, optional) The OnClick script of the button ("Button" type only)
-- @field Enter (function, optional) The OnEnter script of the button
-- @field Leave (function, optional) The OnLeave script of the button
-- @field Tooltip (table) The buttons tooltip
-- @field Value (depending, or function) The controls value or the getter/setter function for values (non "Button" type only)
-- @field UpdateMode (string) The value update mode (OnValueChange, OnFocusLost, Explicit; non "Button" type only)
-- @field Callback (function) The value update callback function (non "Button" type only)
-- @field MinValue (number) The minimum value of the slider for slider type buttons ("Slider" type only)
-- @field MaxValue (number) The maximum value of the slider for slider type buttons ("Slider" type only)
-- @field ValueStep (string) The value step mode of the slider for slider type buttons (INT, or FLOAT; "Slider" type only)
-- @field Items (table) The children dropdown / value collection for Button / DropDown button types ("Button" and "DropDown" type only)
--

---
-- @class table
-- @name Tooltip
-- @description The information (array of lines) about a tooltip (used in field //Tooltip// of dropdown buttons)
-- @field NormalLine (string) Inserting a simple string into the array will create a normal line
-- @field DoubleLine (array[2] of strings) Inserting an array of two strings will create a double line using both tooltip columns
-- @field Seprarator (boolean) Inserting the boolean //true// into the array will create a separator
--

--- The OnClick handler for dropdown buttons (callable via colon call //self:Click("LBUTTON")//)
-- @param self (frame) The actual dropdown button
-- @param key (string) The mouse button the button got clicked with
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.Click(self, key, data)
	catch(self, "Click: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s Click: Invalid data (nil value, table expected)")

	if data.Type == "Label" or data.Enabled == false then
		return
	elseif data.Type == "DropDown" and data.Items then
		local dropdown = {
			{
				Title = TEXT("CANCEL"),
				Icon = "interface/mainmenuframe/mainpopupbutton-icon-close.tga",
				Click = function() DropDown.Back() end,
				Type = "Button",
			},
		}
		for idx, value in pairs(data.Items) do
			table.insert(dropdown, {
				Title = tostring(value),
				Icon = idx == self:GetValue() and "interface/buttons/checkbox-checked.tga" or "",
				Click = function() DropDown.Back() self:UpdateValue(idx, "OnValueChange", data) end,
				Type = "Button",
			})
		end
		DropDown.Set(dropdown)
	elseif data.Type == "CheckButton" then
		local newValue = not self:GetValue()
		self.CheckButton:SetChecked(newValue)
		self:UpdateValue(newValue, "OnValueChange", data)
	elseif data.Type == "Color" then
		if data.Color then
			local r, g, b = unpack(data.Color)
			OpenColorPickerFrameEx({
				titleText = data.Title,
				r = r,
				g = g,
				b = b,
				callbackFuncOkay = function()
					SupressAutoClose = false
					SupressChange = false
					local r, g, b = ColorPickerFrame.r, ColorPickerFrame.g, ColorPickerFrame.b
					self.TexIcon:SetColor(r, g, b)
					self:UpdateValue({r, g, b}, data.UpdateMode, data)
				end,
				callbackFuncUpdate = function()
					local r, g, b = ColorPickerFrame.r, ColorPickerFrame.g, ColorPickerFrame.b
					self.TexIcon:SetColor(r, g, b)
					self:UpdateValue({r, g, b}, "OnValueChange", data)
				end,
				callbackFuncCancel = function()
					SupressAutoClose = false
					SupressChange = false
					local r, g, b = ColorPickerFrame.r, ColorPickerFrame.g, ColorPickerFrame.b
					self.TexIcon:SetColor(r, g, b)
					self:UpdateValue({r, g, b}, data.UpdateMode, data)
				end,
			})
			SupressAutoClose = true
			SupressChange = true
		end
	else
		if data.Type == "Button" and data.Items then
			DropDown.Set(data.Items)
		else
			if data.Click then
				local success, error = pcall(runScript, data.Click, self, key, data)
				if not success then
					bugf("Error while processing OnClick (%s) of Dropdownbutton '%s'!\nError: %s", key, data.Title, tostring(error))
				end
			end
		end
	end
end

--- The OnEnter handler for dropdown buttons (callable via colon call //self:Enter()//)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.Enter(self, data)
	catch(self, "Enter: Invalid button (nil value, table expected)")
	MouseoverButton = self

	data = data or self:GetData()
	catch(data, self.Name.."'s Enter: Invalid data (nil value, table expected)")

	DropDown.StopTimeout()
	if data.Type == "Label" or data.Enabled == false then
		return
	end

	--self.Highlight:Show()
	zzStory.New("awsmDropdown_Highlight"..self:GetID(), {
		{
			_Frame = self.Highlight,
			_Start = function()
				self.Highlight:Show()
			end,
			_Time = 0.2,
			SetAlpha = {
				From = {0},
				To = {1}
			},
		},
	}, 30)
	if data.Type == "Slider" then
		self.Slider:Show()
		self.LabelName:Hide()
		self.LabelValue:Hide()
	elseif data.Type == "EditBox" then
		self.EditBox:Show()
		self.LabelName:Hide()
		self.LabelValue:Hide()
	end

	if data.Enter then
		zzTimer.Add({0.1, 0}, function()
			local success, error = pcall(runScript, data.Enter, self, data)
			if not success then
				bugf("Error while processing OnEnter of Dropdownbutton '%s'!\nError: %s", data.Title, tostring(error))
			end
		end, "awsmDropdown_EnterCooldown"..self:GetID())
	end

	if data.Tooltip then
		zzLibrary.Tooltip(self, data.Title, data.Tooltip, nil, ADDTooltip)
	elseif self.LabelName:IsDrawDot() then
		zzLibrary.Tooltip(self, data.Title, nil, nil, ADDTooltip)
	end
end

--- The OnLeave handler for dropdown buttons (callable via colon call //self:Leave()//)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.Leave(self, data)
	catch(self, "Leave: Invalid button (nil value, table expected)")
	if MouseoverButton == self then
		MouseoverButton = nil
	end

	if self:IsVisible() and DropDown.Frame:IsVisible() then
		data = data or self:GetData()
		catch(data, self.Name.."'s Leave: Invalid data (nil value, table expected)")

		DropDown.StartTimeout()
		if data.Type == "Label" then
			return
		end

		--self.Highlight:Hide()
		zzStory.New("awsmDropdown_Highlight"..self:GetID(), {
			{
				_Frame = self.Highlight,
				_End = function()
					self.Highlight:Hide()
				end,
				_Time = 0.2,
				SetAlpha = {
					From = {1},
					To = {0}
				},
			},
		}, 30)
		if data.Type == "Slider" then
			self.Slider:Hide()
			self.LabelName:Show()
			self.LabelValue:Show()
		elseif data.Type == "EditBox" then
			self.EditBox:Hide()
			self.LabelName:Show()
			self.LabelValue:Show()
		end

		if data.UpdateMode == "OnFocusLost" then
			self:CallUpdate(self:GetValue(), data)
		end
	else
		self.Highlight:Hide()
		self.EditBox:Hide()
		self.Slider:Hide()
		self.LabelName:Show()
		self.LabelValue:Show()
	end

	if data and data.Leave then
		local success, error = pcall(runScript, data.Leave, self, data)
		if not success then
			bugf("Error while processing OnEnter of Dropdownbutton '%s'!\nError: %s", data.Title, tostring(error))
		end
	end

	if ADDTooltipTextLeft1:GetText() then
		zzLibrary.TooltipHide(ADDTooltip)
	end
end

--- Returns a dropdown buttons value (calls a getter function if needed; callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.GetValue(self, data)
	catch(self, "GetValue: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s GetValue: Invalid data (nil value, table expected)")

	if type(data.Value) == "function" then
		return data.Value()
	else
		return data.Value
	end
end

--- Sets a dropdown buttons value (calls a setter function if needed; callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param newValue (depending on button type) The new value to set
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.SetValue(self, newValue, data)
	catch(self, "SetValue: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s SetValue: Invalid data (nil value, table expected)")

	if type(data.Value) == "function" then
		data.Value(newValue)
	else
		data.Value = newValue
	end
	return newValue
end

--- Performs a full dropdown button value update (sets value; updates display; does callback; callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param newValue (depending on button type) The new value to set
-- @param updateMode (string) The context in which the update is called ("OnValueChange", "OnFocusLost", "Explicit")
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.UpdateValue(self, newValue, updateMode, data)
	catch(self, "UpdateValue: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s DrawContent: Invalid data (nil value, table expected)")

	if not SupressValueUpdates and self:GetValue() ~= newValue then
		self:SetValue(newValue)
--		print("Request Callback for button: "..data.Title.."; New value: "..tostring(newValue))
--		print(tostring(data.UpdateMode).." = "..tostring(updateMode).." ("..tostring(data.UpdateMode == updateMode)..")")
		if data.UpdateMode == updateMode then
			self:CallUpdate(newValue, data)
		end
		self:SetValueText(data)

		zzEvent.Trigger("ADD_BUTTON_UPDATED", {
			ButtonFrame = self,
			ButtonData = data,
			NewValue = newValue,
			UpdateMode = updateMode,
		})
	end
end

--- Do the value updated callback (protected call; handles errors; callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param newValue (depending on button type) The new value the callback function gets informed about
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.CallUpdate(self, newValue, data)
	catch(self, "CallUpdate: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s CallUpdate: Invalid data (nil value, table expected)")

	if data.Callback then
		local success, error = pcall(runScript, data.Callback, self, newValue, data)
		if not success then
			bugf("Error while processing update callback (new value: %s) of Dropdownbutton '%s'!\nError: %s", tostring(newValue), data.Title, tostring(error))
		end
	end
end

--- The OnMouseWheel handler of a dropdown button (callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param delta (number) The signed scroll delta (> 0: Upwards; < 0: Downwards)
--
function Button.Scroll(self, delta)
	catch(delta, "Scroll: Invalid scroll delta (nil value, number expected)")
	DropDown.ScrollBy(delta)
end

--- Redraws the whole content of a dropdown button (callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.DrawContent(self, data)
	catch(self, "SetControls: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s DrawContent: Invalid data (nil value, table expected)")

	data.Type = data.Type or "Button"
	data.UpdateMode = data.UpdateMode or "OnValueChange"

	self.LabelName:SetText(data.Title)
	self.TexIcon:SetTexture(data.Icon)

	if data.Separator then
		self.Separator:Show()
	else
		self.Separator:Hide()
	end

	if data.Type == "Label" then
		self.Highlight:Hide()
	elseif MouseoverButton == self then
		self.Highlight:Show()
	end

	self:SetControls(data)
	self:SetValueText(data)
end

--- Initializes the dropdown button controls depending on the buttons type, specified in data (callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.SetControls(self, data)
	catch(self, "SetControls: Invalid button (nil value, table expected)")
	data = data or self:GetData()
	catch(data, self.Name.."'s SetControls: Invalid data (nil value, table expected)")

	if data.Enabled == false then
		self:Disable()
		self.CheckButton:Disable()
		self.Slider:Disable()
		self.EditBox:Disable()
		self.LabelName:SetColor(0.7,0.7,0.7)
		self.LabelValue:SetColor(0.7,0.7,0.7)
		self.TexIcon:SetColor(0.7,0.7,0.7)
	else
		self:Enable()
		self.CheckButton:Enable()
		self.Slider:Enable()
		self.EditBox:Enable()
		self.LabelName:SetColor(1,1,1)
		self.LabelValue:SetColor(1,1,1)
		self.TexIcon:SetColor(1,1,1)
	end

	self.CheckButton:Hide()
	self.Slider:Hide()
	self.EditBox:Hide()

	SupressValueUpdates = true
	if data.Type == "CheckButton" then
		self.CheckButton:Show()
		self.CheckButton:SetChecked(self:GetValue(data))
	elseif data.Type == "Slider" then
		self.Slider:SetMinMaxValues(tonumber(data.MinValue), tonumber(data.MaxValue))
		self.Slider:SetValueStepMode(data.ValueStep or "FLOAT")
		self.Slider:SetValue(tonumber(self:GetValue(data)))
		if self.Highlight:IsVisible() then
			self.Slider:Show()
		end
	elseif data.Type == "EditBox" then
		self.EditBox:SetText(tostring(self:GetValue(data)))
		if self.Highlight:IsVisible() then
			self.EditBox:Show()
		end
	elseif data.Type == "Color" then
		if data.Color then
			self.TexIcon:SetColor(unpack(data.Color))
		end
	end
	SupressValueUpdates = false
end

--- Handles the content redraw of value displaying controls of the dropdown button (callable via colon call)
-- @param self (frame) The actual dropdown button
-- @param data (table, optional) The data that is displayed on the button (gets internally requested if missing)
--
function Button.SetValueText(self, data)
	catch(self, "SetValueText: Invalid button (nil value, table expected)")
	data = data or self:GetData()

	if data then
		if data.Type == "Slider" then
			self.LabelValue:SetText(tostring(self:GetValue()))
		elseif data.Type == "EditBox" then
			self.LabelValue:SetText(tostring(self:GetValue()))
		elseif data.Type == "Label" then
			self.LabelValue:SetText(tostring(self:GetValue() or ""))
		elseif data.Type == "Button" then
			self.LabelValue:SetText(tostring(self:GetValue() or ""))
		elseif data.Type == "DropDown" then
			if data.Items then
				self.LabelValue:SetText(tostring(data.Items[tonumber(self:GetValue())]))
			else
				self.LabelValue:SetText("")
			end
		else
			self.LabelValue:SetText("")
		end
	else
		self.LabelValue:SetText("")
	end
end

--- Returns the data that is displayed on the button depending on button index and scroll offset (callable via colon call)
-- @param self (frame) The actual dropdown button
--
function Button.GetData(self)
	if History[1] then
		return History[1][self:GetID() + DropDown.Scroll:GetValue()]
	end
end

--- Initializes some fields in the dropdown button frametable (enables functions to be called via colon call; important)
-- @param button (frame) The dropdown button to initilize
--
function Button.Initialize(button)
	catch(button, "Initialize: Failed to initialize button (nil value, table expected")
	button:RegisterForClicks("LeftButton", "RightButton")

	button.LabelName = _G[button:GetName().."_Name"]
	button.LabelValue = _G[button:GetName().."_Value"]
	button.TexIcon = _G[button:GetName().."_Icon"]
	button.Separator = _G[button:GetName().."_Separator"]
	button.Highlight = _G[button:GetName().."_Highlight"]
	button.Name = button:GetName()

	button.ButtonClick = Button.Click
	button.ButtonEnter = Button.Enter
	button.ButtonLeave = Button.Leave
	button.DrawContent = Button.DrawContent
	button.GetData = Button.GetData
	button.GetValue = Button.GetValue
	button.SetValue = Button.SetValue
	button.UpdateValue = Button.UpdateValue
	button.SetControls = Button.SetControls
	button.SetValueText = Button.SetValueText
	button.CallUpdate = Button.CallUpdate
	button.Scroll = Button.Scroll
	button.ScrollUp = DropDown.ScrollUp
	button.ScrollDown = DropDown.ScrollDown
end



--- Returns current build number of awsmDropdown
-- @return build (number): The build number
--
function DropDown.GetBuildNumber() return BUILD end

--- Initializes some internal fields with the dropdown frame and children of it (important)
-- @param dropdown (frame) The dropdown frame
--
function DropDown.Initialize(dropdown)
	catch(dropdown, "Initialize: Failed to initialize dropdown (nil value, table expected")

	DropDown.Frame = dropdown
	DropDown.Scroll = _G[dropdown:GetName().."_SliderWrapper_Slider"]
	DropDown.SliderWrapper = _G[dropdown:GetName().."_SliderWrapper"]

	DropDown.Buttons = {}
	for i = 1, 20 do
		local button = _G[dropdown:GetName().."_ScrollFrame_Content_Button"..i]
		DropDown.Buttons[i] = button
	end
end

--- Reanchors the awsmDropdown frame
-- @param frame (table/string): The frame to anchor the dropdown to
-- @param anchorPoint (string): The anchor point
-- @param relativePoint (string): The relative point
-- @param offsetX (number): Horizontal offset
-- @param offsetY (number): Vertical offset
--
function DropDown.Anchor(frame, relativePoint, anchorPoint, offsetX, offsetY)
	catch(type(frame) == "string" or type(frame) == "table", "Anchor: Must give a frame to anchor to (string or table expected)")
	catch(type(anchorPoint) == "string", "Anchor: Must give a anchor point (string expected)")
	catch(type(relativePoint) == "string", "Anchor: Must give a relative point (string expected)")
	catch(type(offsetX) == "number", "Anchor: aoffsetX must be  number, got: "..type(offsetX))
	catch(type(offsetY) == "number", "Anchor: offsetX must be a number, got: "..type(offsetY))

	local eventArgs = {
		Cancel = false,
		Anchor = { frame, relativePoint, anchorPoint, offsetX, offsetY },
	}
	zzEvent.Trigger("ADD_ANCHOR_CHANGING", eventArgs)

	if eventArgs.Cancel == true then
		return
	end

	DropDown.Frame:ClearAllAnchors()
	DropDown.Frame:SetAnchor(anchorPoint, relativePoint, frame, offsetX, offsetY)
end

--- Checks whether the dropdown frame is visible
-- @return isVisible (bool): Visibility of the dropdown frame
--
function DropDown.IsVisible()
	return DropDown.Frame:IsVisible()
end

--- Returns the current anchor of the dropdown frame
-- @return anchorPoint (string): Anchoring point on anchored frame
-- @return relativePoint (string): Relative anchoring point on relative frame
-- @return relativeTo (): Relative frame
-- @return offsetX (number): X offset
-- @return offsetY (number): Y offset
--
function DropDown.GetAnchor()
	return DropDown.Frame:GetAnchor()
end

--- Shows the dropdown frame and displays the given dropdown (more information in special page about this)
-- @param dropdown (table): The dropdown to display
--
function DropDown.Open(dropdown)
	if not SupressChange then
		catch(dropdown, "Open: No dropdown to open (nil value, table expected)")

		local eventArgs = {
			DropDown = dropdown,
			Cancel = false,
			Anchor = { DropDown.GetAnchor() },
		}
		zzEvent.Trigger("ADD_DROPDOWN_OPENING", eventArgs)

		if eventArgs.Cancel == true then
			return
		end

		DropDown.Set(dropdown)
		--DropDown.Frame:Show()
		zzStory.New("awsmDropdownFrame", {
			{
				_Frame = DropDown.Frame,
				_Start = function()
					DropDown.Frame:Show()
					PlaySoundByPath("sound/interface/ui_dropdown_open.mp3")
				end,
				_End = function()
					eventArgs.Cancel = nil
					zzEvent.Trigger("ADD_DROPDOWN_OPENED", eventArgs)
				end,
				_Time = 0.125,
				SetAlpha = {
					From = {0},
					To = {1}
				},
				SetScale = {
					From = {0.8},
					To = {1}
				},
			},
		}, 60)
	end
end

--- Hides the dropdown frame and resets the history
--
function DropDown.Close()
	--DropDown.Frame:Hide()
	--DropDown.ResetHistory()

	if not SupressChange then
		local eventArgs = {
			DropDown = History[1],
			Cancel = false,
			Anchor = { DropDown.GetAnchor() },
		}
		zzEvent.Trigger("ADD_DROPDOWN_CLOSING", eventArgs)

		if eventArgs.Cancel == true then
			return
		end
		
		zzStory.New("awsmDropdownFrame", {
			{
				_Frame = DropDown.Frame,
				_End = function()
					DropDown.Frame:Hide()
					DropDown.Frame:SetHeight(16)
					DropDown.ResetHistory()

					eventArgs.Cancel = nil
					zzEvent.Trigger("ADD_DROPDOWN_CLOSED", eventArgs)
				end,
				_Time = 0.1,
				SetAlpha = {
					From = {1},
					To = {0}
				},
				SetScale = {
					From = {1},
					To = {0.8}
				},
			},
		}, 60)
	end
end

--- Toggles the dropdown frame and displays the given dropdown on show, optionally reanchoring it
-- @param dropdown (table): The dropdown to display on show
-- @param frame (table/string, optional): The frame to anchor the dropdown to
-- @param anchorPoint (string, optional): The anchor point
-- @param relativePoint (string, optional): The relative point
-- @param offsetX (number, optional): Horizontal offset
-- @param offsetY (number, optional): Vertical offset
--
function DropDown.Toggle(dropdown, frame, relativePoint, anchorPoint, offsetX, offsetY)
	if not SupressChange then
		if DropDown.IsVisible() then
			DropDown.Close()
		else
			if frame and relativePoint and anchorPoint and offsetX and offsetY then
				DropDown.Anchor(frame, relativePoint, anchorPoint, offsetX, offsetY)
			end
			DropDown.Open(dropdown)
		end
	end
end

--- Sets inserts the given dropdown to the history and displays it
-- @param dropdown (table): The dropdown to display
--
function DropDown.Set(dropdown, overwrite)
	if not SupressChange then
		local eventArgs = {
			OldDropDown = History[1],
			NewDropDown = dropdown,
			Cancel = false,
			Anchor = { DropDown.GetAnchor() },
		}
		zzEvent.Trigger("ADD_DROPDOWN_CHANGING", eventArgs)

		if eventArgs.Cancel == true then
			return
		end

		if dropdown then
			if overwrite == true then
				History[1] = dropdown
			else
				table.insert(History, 1, dropdown)
			end
		end

		catch(History[1], "Set: Can't read dropdown from history (nil value, table expected)")


		local visible, total = DropDown.GetVisibleItems()
		
		if DropDown.Scroll then
			DropDown.Scroll:SetMinMaxValues(0, total - visible)
			DropDown.ScrollTo(History[1].Scroll or 0)
		end

		DropDown.DrawItems()

		eventArgs.Cancel = nil
		zzEvent.Trigger("ADD_DROPDOWN_CHANGED", eventArgs)
	end
end

--- Gets the currently displayed dropdown
-- @return dropdown (table): The current dropdown
--
function DropDown.Get()
	catch(History[1], "Get: Can't get currentdropdown. History is empty.")
	return History[1]
end

--- Sets the dropdown to the previous one in the dropdown history
--
function DropDown.Back()
	if not SupressChange then
		catch(History[1], "Back: Can't read dropdown from history (nil value, table expected)")
		catch(History[2], "Back: No dropdown in history to go back to (nil value, table expected)")

		table.remove(History, 1)
		DropDown.Set()
	end
end

--- Resets the dropdown history
--
function DropDown.ResetHistory()
	if not SupressChange then
		History = {}
	end
end

--- Gets the number of visible items in the dropdown
-- @return visible (number): Number of visible dropdown items
-- @return total (number): Total number of dropdown items
--
function DropDown.GetVisibleItems()
	catch(History[1], "GetVisibleItems: Can't read dropdown from history (nil value, table expected)")

	local numItems = #(History[1])
	return (numItems > 20 and 20 or numItems), numItems
end

--- Gets a button from the current dropdown by it's title
-- @param title (string): The buttons title to search for
-- @return button (table): The buttons data
-- @return index (number): The buttons index
--
function DropDown.GetItemByTitle(title)
	catch(History[1], "GetItemByName: Can't read dropdown from history (nil value, table expected)")
	for idx, button in pairs(History[1]) do
		if button.Title == title then
			return button, idx
		end
	end
end

--- Redraws all dropdown items and sets dropdown frame dimensions
--
function DropDown.DrawItems()
	catch(DropDown.Buttons, "ScrollDown: Failure accessing dropdown buttons (nil value)")

	local visible, count = DropDown.GetVisibleItems()
	for i = 1, visible do
		DropDown.Buttons[i]:Show()
		DropDown.Buttons[i]:DrawContent()
	end
	for i = visible + 1, 20 do
		DropDown.Buttons[i]:Hide()
	end

	if count > 20 then
		DropDown.Scroll:Show()
		DropDown.Scroll:SetWidth(16)
		DropDown.SliderWrapper:Show()
		DropDown.SliderWrapper:SetWidth(16)
	else
		DropDown.Scroll:Hide()
		DropDown.Scroll:SetWidth(0)
		DropDown.SliderWrapper:Hide()
		DropDown.SliderWrapper:SetWidth(0)
	end

	local oldHeight = DropDown.Frame:GetHeight()
	local newHeight = 12 + DropDown.Buttons[1]:GetHeight() * visible + (visible == 0 and 4 or 0)

	--DropDown.Frame:SetHeight(12 + DropDown.Buttons[1]:GetHeight() * visible)
	if math.abs(oldHeight - newHeight) > 0.0001 then
		zzStory.New("awsmDropdownDraw", {
			{
				_Frame = DropDown.Frame,
				_Time = 0.01 * (#DropDown.Buttons),
				_End = function()
					DropDown.Frame:SetHeight(newHeight)
					DropDown.KeepInScreen()
				end,
				SetHeight = {
					From = {oldHeight},
					To = {newHeight}
				},
			},
		}, 60)
	end
end

local ConfigBuilder = {
	Color = function(setting, index, title, configTable, applyFunctions)
		local r, g, b = unpack(configTable[setting.Name])
		return {
			Title = title,
			Icon = "interface/common/block-fill.tga",
			Tooltip = setting.Description,
			Type = "Color",
			Value = function() return configTable[setting.Name] end,
			UpdateMode = setting.UpdateMode or "OnValueChange",
			Color = {r, g, b},
			Callback = function(frame, newValue, button)
				if configTable[setting.Name][1] ~= newValue[1] or
						configTable[setting.Name][2] ~= newValue[2] or
						configTable[setting.Name][3] ~= newValue[3]
				then
					configTable[setting.Name] = newValue;
					local success, error = pcall(runScript, applyFunctions[setting.Name], frame, newValue, button)
					if not success then
						bugf("Error while processing apply script of option '%s'!|r\nError: %s", tostring(setting.Name), tostring(error))
					end
				end
			end
		}
	end,
	Action = function(setting, index, title, configTable, applyFunctions)
		return {
			Title = title,
			Icon = setting.Icon or "interface/mainmenuframe/mainpopupbutton-icon-system.tga",
			Tooltip = setting.Description,
			Type = "Button",
			Click = setting.Click,
		}
	end,
	CheckButton = function(setting, index, title, configTable, applyFunctions)
		return {
			Title = title,
			Icon = "",
			Tooltip = setting.Description,
			Type = "CheckButton",
			Value = function() return configTable[setting.Name] end,
			UpdateMode = setting.UpdateMode or "OnValueChange",
			Callback = function(frame, newValue, button)
--				print("Callback for option: "..setting.Name.."; New value: "..tostring(newValue))
				if configTable[setting.Name] ~= newValue then
					configTable[setting.Name] = newValue;
					local success, error = pcall(runScript, applyFunctions[setting.Name], frame, newValue, button)
					if not success then
						bugf("Error while processing apply script of option '%s'!|r\nError: %s", tostring(setting.Name), tostring(error))
					end
				end
			end,
		}
	end,
	EditBox = function(setting, index, title, configTable, applyFunctions)
		return {
			Title = title,
			Icon = "interface/mainmenuframe/mainpopupbutton-icon-system.tga",
			Tooltip = setting.Description,
			Type = "EditBox",
			Value = function() return configTable[setting.Name] end,
			UpdateMode = setting.UpdateMode or "OnFocusLost",
			Callback = function(frame, newValue, button)
--				print("Callback for option: "..setting.Name.."; New value: "..tostring(newValue))
				if configTable[setting.Name] ~= newValue then
					configTable[setting.Name] = newValue;
					local success, error = pcall(runScript, applyFunctions[setting.Name], frame, newValue, button)
					if not success then
						bugf("Error while processing apply script of option '%s'!|r\nError: %s", tostring(setting.Name), tostring(error))
					end
				end
			end,
		}
	end,
	Slider = function(setting, index, title, configTable, applyFunctions)
		return {
			Title = title,
			Icon = "interface/mainmenuframe/mainpopupbutton-icon-system.tga",
			Tooltip = setting.Description,
			Type = "Slider",
			Value = function() return configTable[setting.Name] end,
			MinValue = setting.MinValue or 0,
			MaxValue = setting.MaxValue or 1,
			ValueStep = setting.ValueStep or "INT",
			UpdateMode = setting.UpdateMode or "OnValueChange",
			Callback = function(frame, newValue, button)
--				print("Callback for option: "..setting.Name.."; New value: "..tostring(newValue))
				if configTable[setting.Name] ~= newValue then
					configTable[setting.Name] = newValue;
					local success, error = pcall(runScript, applyFunctions[setting.Name], frame, newValue, button)
					if not success then
						bugf("Error while processing apply script of option '%s'!|r\nError: %s", tostring(setting.Name), tostring(error))
					end
				end
			end,
		}
	end,
	DropDown = function(setting, index, title, configTable, applyFunctions)
		return {
			Title = title,
			Icon = "interface/mainmenuframe/mainpopupbutton-icon-system.tga",
			Tooltip = setting.Description,
			Type = "DropDown",
			Value = function() return configTable[setting.Name] end,
			Items = setting.Items,
			UpdateMode = setting.UpdateMode or "OnValueChange",
			Callback = function(frame, newValue, button)
--				print("Callback for option: "..setting.Name.."; New value: "..tostring(newValue))
				if configTable[setting.Name] ~= newValue then
					configTable[setting.Name] = newValue;
					local success, error = pcall(runScript, applyFunctions[setting.Name], frame, newValue, button)
					if not success then
						bugf("Error while processing apply script of option '%s'!|r\nError: %s", tostring(setting.Name), tostring(error))
					end
				end
			end,
		}
	end,
}
--- Automatically builds and returns a configuration dropdown from settings (more information in special page about this)
-- @param settings (table) The table containing information about the available settings
-- @param locales (table) The table containing localized setting titles (indexed with setting name)
-- @param configTable (table) The table holding the configuration variables (the one you call //SaveVariables// with)
-- @param applyFunctions (table) The table containing callback functions applying changes made by the user (indexed with setting name)
-- @param autoSet (boolean) Whether to automatically show the generated dropdown
-- @return dropdown (table) The generated configuration dropdown
--
function DropDown.BuildConfig(settings, locales, configTable, applyFunctions, autoSet)
	if type(configTable) == "string" then
		configTable = zzTable.GetSubValue(configTable)
	end
	if type(configTable) ~= "table" then
		throw("Invalid config variables table!", 3)
	end

	if type(applyFunctions) == "string" then
		applyFunctions = zzTable.GetSubValue(applyFunctions)
	end
	if type(applyFunctions) ~= "table" then
		throw("Invalid apply functions table!", 3)
	end

	local dropdown = {
		{
			Title = "< "..TEXT("UI_GUILDHOUSEBUILDING_BACK"),
			Icon = "interface/mainmenuframe/mainpopupbutton-icon-close.tga",
			Click = function(button, key, data) awsmDropdown.Back() end,
			Type = "Button",
		},
	}
	for i, setting in ipairs(settings) do
		if setting.Visible ~= false then
			catch(setting.Name, "Can't generate settings dropdown without setting name at idx "..i)
			catch(setting.Type, "Can't generate settings dropdown without setting type at idx "..i)
			local title = locales[string.upper(setting.Name)] or string.upper(setting.Name)
			local builder = ConfigBuilder[setting.Type]
			catch(builder, "Can't build settings for type "..tostring(setting.Type).." at idx "..i)

			table.insert(dropdown, builder(setting, i, title, configTable, applyFunctions))
		end
	end

	if autoSet == true then
		DropDown.Set(dropdown)
	end
	return dropdown
end

--- Starts the timeout that hides the dropdown frame after 2 seconds
--
function DropDown.StartTimeout()
	if not SupressAutoClose and not SupressChange then
		local eventArgs = {
			DropDown = History[1],
			Cancel = false,
			Anchor = { DropDown.GetAnchor() },
		}
		zzEvent.Trigger("ADD_TIMEOUT_STARTING", eventArgs)

		if eventArgs.Cancel == true then
			return
		end

		zzTimer.Add({2, 0}, function()
			DropDown.Close()
			zzLibrary.TooltipHide(ADDTooltip)
		end, "awsmDropdown_TimeoutClose")

		eventArgs.Cancel = nil
		zzEvent.Trigger("ADD_TIMEOUT_STARTED", eventArgs)
	end
end

--- Stops the timeout that hides the dropdown frame after 2 seconds
--
function DropDown.StopTimeout()
	local eventArgs = {
		DropDown = History[1],
		Cancel = false,
		Anchor = { DropDown.GetAnchor() },
	}
	zzEvent.Trigger("ADD_TIMEOUT_STOPPING", eventArgs)

	if eventArgs.Cancel == true then
		return
	end

	zzTimer.Remove("awsmDropdown_TimeoutClose")

	eventArgs.Cancel = nil
	zzEvent.Trigger("ADD_TIMEOUT_STOPPED", eventArgs)
end

--- Scrolls the dropdown up or down, depending on delta
-- @param delta (number) The scroll delta usually available in OnMouseWheel handlers (positive = up)
--
function DropDown.ScrollBy(delta)
	if delta > 0 then
		DropDown.ScrollUp()
	else
		DropDown.ScrollDown()
	end
end

--- Scrolls up by one button
--
function DropDown.ScrollUp()
	DropDown.ScrollTo((DropDown.Scroll:GetValue() or 1) - 1)
end

--- Scrolls down by one button
--
function DropDown.ScrollDown()
	DropDown.ScrollTo((DropDown.Scroll:GetValue() or 0) + 1)
end

--- Scrolls to the given offset
-- @param offset (number): The offset to scroll to
--
function DropDown.ScrollTo(offset)
	catch(History[1], "ScrollTo: Can't read dropdown from history (nil value, table expected)")

	if offset > DropDown.Scroll:GetMaxValue() or
	   offset < DropDown.Scroll:GetMinValue() or
	   offset == DropDown.Scroll:GetValue()
	then
		return
	end

	History[1].Scroll = offset
	DropDown.Scroll:SetValue(offset)

	DropDown.DrawItems()
end

--- Ensures that the whole Dropdown is located in the visible screen space
--
function DropDown.KeepInScreen()
	local anchorPoint, relativePoint, relativeTo, offsetX, offsetY = DropDown.Frame:GetAnchor()
	offsetX = offsetX or 0
	offsetY = offsetY or 0

	local uiScale = GetUIScale();
	local posTop = DropDown.Frame:GetTop()
	local posRight = DropDown.Frame:GetRight()
	local posBottom = DropDown.Frame:GetBottom()
	local posLeft = DropDown.Frame:GetLeft()
	local screenWidth, screenHeight = GetScreenWidth(), GetScreenHeight()

	local overflowTop = -posTop
	local overflowLeft = -posLeft
	local overflowBottom = posBottom - screenHeight
	local overflowRight = posRight - screenWidth

	if overflowTop > 0 then offsetY = offsetY + overflowTop/uiScale end
	if overflowLeft > 0 then offsetX = offsetX + overflowLeft/uiScale end
	if overflowBottom > 0 then offsetY = offsetY - overflowBottom/uiScale end
	if overflowRight > 0 then offsetX = offsetX - overflowRight/uiScale end

	--DropDown.Frame:SetPos(posLeft, posTop)
	DropDown.Frame:ClearAllAnchors()
	DropDown.Frame:SetAnchor(anchorPoint or "TOP", relativePoint or "CENTER", relativeTo or "UIParent", offsetX, offsetY)
end